// This mode was derived from commonlisp mode

module.exports = function (CodeMirror) {
    "use strict";

    CodeMirror.defineMode("kiniro", function (config) {
        var specialForm = /^(begin|apply|def|let|if|cond|throw|try|set|par|async|await)$/;
        var numLiteral = /[-+]?[0-9]*\.?[0-9]+/;
        var symbol = /[a-zA-Z!?+\->\/_=\*<][a-zA-Z0-9!?+\->\/\\_=\*<]*/;
        var type;
        const brackets = {
            ')': '(',
            ']': '[',
            '}': '{'
        };

        function readSym (stream) {
            stream.match(symbol);
            return stream.current();
        }

        function base (stream, state) {
            if (stream.eatSpace()) {type = "ws"; return null;}
            if (stream.match(numLiteral)) return "number";
            if (stream.match('//')) { stream.skipToEnd(); return "comment"; }
            if (stream.match(/(->|\\|\$)/)) { return "sugar"; }
            var ch = stream.next();

            if (ch == '"') return (state.tokenize = inString)(stream, state);
            else if (ch == "(" || ch == "[" || ch == '{') { type = "open";
                                               state.brackets.push(ch);
                                               return `bracket_${(state.brackets.length - 1) % state.parenth_max_depth}`; }
            else if (ch == ")" || ch == "]" || ch == '}') { type = "close";
                                               if (state.brackets.length == 0 || state.brackets[state.brackets.length - 1] != brackets[ch]) {
                                                   return "bracket_unmatched";
                                               }
                                               state.brackets.pop();
                                               return `bracket_${state.brackets.length % state.parenth_max_depth}`; }
            else if (ch == ";") { stream.skipToEnd(); type = "ws"; return "comment"; }
            else if (ch == "/") {
                var ch = stream.next();
                /*if (ch == "(") { type = "open"; return "bracket"; }
                else if (/[+\-=\.']/.test(ch)) return null;
                else if (/\d/.test(ch) && stream.match(/^\d*#/)) return null;
                else */
                if (ch == "*") return (state.tokenize = inComment)(stream, state);
                /* else if (ch == ":") { readSym(stream); return "meta"; }
                else if (ch == "\\") { stream.next(); readSym(stream); return "string-2" } */
                else return "error";
            } else {
                var name = readSym(stream);
                type = "symbol";
                if (state.lastType == "open" && (specialForm.test(name))) return "keyword";
                return "variable";
            }
        }

        function inString (stream, state) {
            var escaped = false, next;
            while ((next = stream.next())) {
                if (next == '"' && !escaped) { state.tokenize = base; break; }
                escaped = !escaped && next == "\\";
            }
            return "string";
        }

        function inComment (stream, state) {
            var next, last;
            while ((next = stream.next())) {
                if (next == "/" && last == "*") { state.tokenize = base; break; }
                last = next;
            }
            type = "ws";
            return "comment";
        }

        return {
            startState: function () {
                return {ctx: {prev: null, start: 0, indentTo: 0}, depth: 0, brackets: [],  parenth_max_depth: 10, lastType: null, tokenize: base};
            },

            token: function (stream, state) {
                if (stream.sol() && typeof state.ctx.indentTo != "number")
                    state.ctx.indentTo = state.ctx.start + 1;

                type = null;
                var style = state.tokenize(stream, state);
                if (type != "ws") {
                    if (state.ctx.indentTo == null) {
                        state.ctx.indentTo = "next";
                    } else if (state.ctx.indentTo == "next") {
                        state.ctx.indentTo = stream.column();
                    }
                    state.lastType = type;
                }
                if (type == "open") state.ctx = {prev: state.ctx, start: stream.column(), indentTo: null};
                else if (type == "close") state.ctx = state.ctx.prev || state.ctx;
                return style;
            },

            indent: function (state, _textAfter) {
                var i = state.ctx.indentTo;
                return typeof i == "number" ? i : state.ctx.start + 1;
            },

            /* used by plugins */
            closeBrackets: {pairs: "()[]{}\"\""},
            lineComment: ";;",
            blockCommentStart: "/*",
            blockCommentEnd: "*/"
        };
    });
};
